import asyncio
import threading
import time

from newt import Queue

def test_threaded_in_executor_with_coroutine():
    queue = Queue()

    loop = asyncio.get_event_loop()

    def threaded(q):
        for i in range(100):
            q.put(i)
        q.join()


    async def coroutine(q):
        for i in range(100):
            assert await q.get() == i
            q.task_done()


    async def main():
        future = loop.run_in_executor(None, threaded, queue.sync_q)
        await coroutine(queue.async_q)
        await future

        queue.close()
        await queue.wait_closed()

    loop.run_until_complete(main())


def test_threading_with_coroutine():
    queue = Queue()
    loop = asyncio.get_event_loop()

    def threaded(q):
        for i in range(100):
            q.put(i)
        q.join()

    async def coroutine(q):
        for i in range(100):
            got = await q.get()
            assert got == i
            q.task_done()

    async def main():
        await coroutine(queue.async_q)
        queue.close()
        await queue.wait_closed()

    t = threading.Thread(target=threaded, args=(queue.sync_q,))
    t.start()

    time.sleep(.1)

    loop.run_until_complete(main())



from typing import (
    TypeVar,
    Optional,
    Callable
)
import asyncio


T = TypeVar('T')
OptInt = Optional[int]


def lazy_property(fn: Callable[..., T]):
    """Lazily initialize a property
    """

    name = fn.__name__
    key = f'{name}_'

    def helper(self, *args, **kwargs) -> T:
        value = getattr(self, key, None)

        if value is None:
            value = fn(self, *args, **kwargs)
            setattr(self, key, value)

        return value

    return property(helper)


def has_loop(fn):
    """
    Only execute fn when there is a loop
    """
    def helper(self, *args, **kwargs) -> None:
        if self._loop_ is None:
            return

        fn(self, *args, **kwargs)

    return helper


def check_closing(fn: Callable[..., T]):
    def helper(self, *args, **kwargs) -> T:
        if self._parent._closing:
            raise RuntimeError('modification of closed queue is forbidden')

        return fn(self, *args, **kwargs)

    return helper


get_running_loop = asyncio.get_event_loop \
    if getattr(asyncio, 'get_running_loop', None) is None \
    else asyncio.get_running_loop



import asyncio
from asyncio import (
    AbstractEventLoop,
    Future
)
import threading
from abc import ABC, abstractmethod

from typing import (
    Generic,
    Any,
    Set,
    Callable,
    Optional,
    Awaitable
)

from .common import (
    T,
    lazy_property,
    has_loop,
    get_running_loop
)


class AbstractQueue(Generic[T], ABC):
    _loop_: Optional[AbstractEventLoop]
    _pending: Set[Future]

    def __init__(self, maxsize: int = 0) -> None:
        self._loop_ = None
        self._maxsize = maxsize

        self._init(maxsize)

        self._unfinished_tasks = 0

        sync_mutex = threading.Lock()
        self._sync_mutex = sync_mutex

        self._sync_not_empty = threading.Condition(sync_mutex)
        self._sync_not_full = threading.Condition(sync_mutex)
        self._all_tasks_done = threading.Condition(sync_mutex)

        try:
            async_mutex = asyncio.Lock()
        except RuntimeError as e:
            raise RuntimeError(f'''{e}

This is usually not an issue of newt.
You should make sure there is a event loop in the current thread, especially after running `asyncio.run()`.

Check https://github.com/kaelzhang/python-newt for details.''')

        self._async_mutex = async_mutex

        self._async_not_empty = asyncio.Condition(async_mutex)
        self._async_not_full = asyncio.Condition(async_mutex)
        self._finished = asyncio.Event()
        self._finished.set()

        self._closing = False
        self._pending = set()

    @lazy_property
    def _loop(self) -> asyncio.AbstractEventLoop:
        return get_running_loop()

    def close(self) -> None:
        with self._sync_mutex:
            self._closing = True
            for fut in self._pending:
                fut.cancel()

    async def wait_closed(self) -> None:
        # should be called from loop after close().
        # Nobody should put/get at this point,
        # so lock acquiring is not required
        if not self._closing:
            raise RuntimeError("waiting for non-closed queue")
        # give execution chances for the task-done callbacks
        # of async tasks created inside
        # _notify_async_not_empty, _notify_async_not_full
        # methods.
        await asyncio.sleep(0)
        if not self._pending:
            return
        await asyncio.wait(self._pending)

    @property
    def closed(self) -> bool:
        return self._closing and not self._pending

    @property
    def maxsize(self) -> int:
        return self._maxsize

    def _put_internal(self, item: T) -> None:
        self._put(item)
        self._unfinished_tasks += 1
        self._finished.clear()

    # Override these methods to implement other queue organizations
    # --------------------------------------------------------------

    # (e.g. stack or priority queue).
    # These will only be called with appropriate locks held

    @abstractmethod
    def _init(self, maxsize: int) -> None:
        ...

    @abstractmethod
    def _qsize(self) -> int:
        ...

    @abstractmethod
    def _put(self, item: T) -> None:
        """Put a new item in the queue
        """
        ...

    @abstractmethod
    def _get(self) -> T:
        """Get an item from the queue
        """
        ...

    # Utilities for async queue to notify sync queue
    # --------------------------------------------------------------

    # This method is always called in a event loop,
    # so we do not need to check loop initialization
    def _notify_sync_not_empty(self) -> None:
        def f() -> None:
            with self._sync_mutex:
                self._sync_not_empty.notify()

        self._loop.run_in_executor(None, f)

    def _notify_sync_not_full(self) -> None:
        def f() -> None:
            with self._sync_mutex:
                self._sync_not_full.notify()

        fut = self._loop.run_in_executor(None, f)
        fut.add_done_callback(self._pending.discard)  # type: ignore
        self._pending.add(fut)  # type: ignore

    # Utilities for sync queue to notify async queue
    # --------------------------------------------------------------

    # If loop is not initialized, then do nothing
    @has_loop
    def _notify_async_not_empty(self, *, threadsafe: bool) -> None:
        async def f() -> None:
            async with self._async_mutex:
                self._async_not_empty.notify()

        def task_maker() -> None:
            task = self._loop.create_task(f())
            task.add_done_callback(self._pending.discard)
            self._pending.add(task)

        if threadsafe:
            self._call_soon_threadsafe(task_maker)
        else:
            self._call_soon(task_maker)

    @has_loop
    def _notify_async_not_full(self, *, threadsafe: bool) -> None:
        async def f() -> None:
            async with self._async_mutex:
                self._async_not_full.notify()

        def task_maker() -> None:
            task = self._loop.create_task(f())
            task.add_done_callback(self._pending.discard)
            self._pending.add(task)

        if threadsafe:
            self._call_soon_threadsafe(task_maker)
        else:
            self._call_soon(task_maker)

    def _call_soon_threadsafe(
        self,
        callback: Callable[..., None],
        *args: Any
    ) -> None:
        try:
            self._loop.call_soon_threadsafe(callback, *args)
        except RuntimeError:
            pass

    def _call_soon(
        self,
        callback: Callable[..., None],
        *args: Any
    ) -> None:
        if not self._loop.is_closed():
            self._loop.call_soon(callback, *args)



from typing import Generic
from asyncio import QueueEmpty
from asyncio import QueueFull

from .common import (
    T,
    check_closing
)
from .queue import AbstractQueue


class AsyncQueueProxy(Generic[T]):
    """Create a queue object with a given maximum size.

    If maxsize is <= 0, the queue size is infinite.
    """

    def __init__(
        self,
        parent: AbstractQueue[T]
    ) -> None:
        self._parent = parent

    @property
    def closed(self) -> bool:
        return self._parent.closed

    def qsize(self) -> int:
        """Number of items in the queue."""
        return self._parent._qsize()

    @property
    def unfinished_tasks(self) -> int:
        """Return the number of unfinished tasks."""
        return self._parent._unfinished_tasks

    @property
    def maxsize(self) -> int:
        """Number of items allowed in the queue."""
        return self._parent._maxsize

    @check_closing
    async def put(self, item: T) -> None:
        """Put an item into the queue.

        Put an item into the queue. If the queue is full, wait until a free
        slot is available before adding item.

        This method is a coroutine.
        """

        async with self._parent._async_not_full:
            self._parent._sync_mutex.acquire()
            locked = True
            try:
                if self._parent._maxsize > 0:
                    do_wait = True
                    while do_wait:
                        do_wait = (
                            self._parent._qsize() >= self._parent._maxsize
                        )
                        if do_wait:
                            locked = False
                            self._parent._sync_mutex.release()
                            await self._parent._async_not_full.wait()
                            self._parent._sync_mutex.acquire()
                            locked = True

                self._parent._put_internal(item)
                self._parent._async_not_empty.notify()
                self._parent._notify_sync_not_empty()
            finally:
                if locked:
                    self._parent._sync_mutex.release()

    @check_closing
    def put_nowait(self, item: T) -> None:
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """

        with self._parent._sync_mutex:
            if self._parent._maxsize > 0:
                if self._parent._qsize() >= self._parent._maxsize:
                    raise QueueFull

            self._parent._put_internal(item)
            self._parent._notify_async_not_empty(threadsafe=False)
            self._parent._notify_sync_not_empty()

    @check_closing
    async def get(self) -> T:
        """Remove and return an item from the queue.

        If queue is empty, wait until an item is available.

        This method is a coroutine.
        """

        async with self._parent._async_not_empty:
            self._parent._sync_mutex.acquire()
            locked = True
            try:
                do_wait = True
                while do_wait:
                    do_wait = self._parent._qsize() == 0

                    if do_wait:
                        locked = False
                        self._parent._sync_mutex.release()
                        await self._parent._async_not_empty.wait()
                        self._parent._sync_mutex.acquire()
                        locked = True

                item = self._parent._get()
                self._parent._async_not_full.notify()
                self._parent._notify_sync_not_full()
                return item
            finally:
                if locked:
                    self._parent._sync_mutex.release()

    def get_nowait(self) -> T:
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """

        with self._parent._sync_mutex:
            if self._parent._qsize() == 0:
                raise QueueEmpty

            item = self._parent._get()
            self._parent._notify_async_not_full(threadsafe=False)
            self._parent._notify_sync_not_full()
            return item

    @check_closing
    def task_done(self) -> None:
        """Indicate that a formerly enqueued task is complete.

        Used by queue consumers. For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items have
        been processed (meaning that a task_done() call was received for every
        item that had been put() into the queue).

        Raises ValueError if called more times than there were items placed in
        the queue.
        """

        with self._parent._all_tasks_done:
            if self._parent._unfinished_tasks <= 0:
                raise ValueError('task_done() called too many times')
            self._parent._unfinished_tasks -= 1
            if self._parent._unfinished_tasks == 0:
                self._parent._finished.set()
                self._parent._all_tasks_done.notify_all()

    async def join(self) -> None:
        """Block until all items in the queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer calls task_done() to
        indicate that the item was retrieved and all work on it is complete.
        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        while True:
            with self._parent._sync_mutex:
                if self._parent._unfinished_tasks == 0:
                    break
            await self._parent._finished.wait()



from heapq import heappop, heappush
from typing import (
    List,
    Deque
)

from collections import deque

from .common import (
    T,
    lazy_property
)
from .queue import AbstractQueue
from .proxy_sync import SyncQueueProxy
from .proxy_async import AsyncQueueProxy

__all__ = (
    'Queue',
    'PriorityQueue',
    'LifoQueue'
)


# The first beta version
__version__ = '0.0.2'


class _AbstractQueue(AbstractQueue[T]):
    @lazy_property
    def sync_q(self) -> SyncQueueProxy[T]:
        return SyncQueueProxy(self)

    @lazy_property
    def async_q(self) -> AsyncQueueProxy[T]:
        return AsyncQueueProxy(self)


class Queue(_AbstractQueue[T]):
    _queue: Deque

    def _init(self, maxsize: int) -> None:
        self._queue = deque()

    def _qsize(self) -> int:
        return len(self._queue)

    # Put a new item in the queue
    def _put(self, item: T) -> None:
        self._queue.append(item)

    # Get an item from the queue
    def _get(self) -> T:
        return self._queue.popleft()


class PriorityQueue(_AbstractQueue[T]):
    """Variant of Queue that retrieves open entries in priority order
    (lowest first).

    Entries are typically tuples of the form:  (priority number, data).

    """

    _heap_queue: List[T]

    def _init(self, maxsize: int) -> None:
        self._heap_queue = []

    def _qsize(self) -> int:
        return len(self._heap_queue)

    def _put(self, item: T) -> None:
        heappush(self._heap_queue, item)

    def _get(self) -> T:
        return heappop(self._heap_queue)


class LifoQueue(_AbstractQueue[T]):
    """Variant of Queue that retrieves most recently added entries first."""

    _queue: Deque

    def _init(self, maxsize: int) -> None:
        self._queue = deque()

    def _qsize(self) -> int:
        return len(self._queue)

    def _put(self, item: T) -> None:
        self._queue.append(item)

    def _get(self) -> T:
        return self._queue.pop()



import os
from pathlib import Path
from setuptools import setup

from newt import __version__

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(Path(__file__).parent / 'docs' / 'README.md').read()

def read_requirements(filename):
    with open(filename) as f:
        return f.read().splitlines()

settings = dict(
    name = 'newt',
    packages = ['newt'],
    version = __version__,
    author = 'kaelzhang',
    author_email = '',
    description = ('Thread-safe, mixed sync-async queue for Python'),
    license = 'MIT',
    keywords = 'newt',
    url = 'https://github.com/kaelzhang/python-newt',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    install_requires=read_requirements('requirements.txt'),
    tests_require=read_requirements('test-requirements.txt'),
    classifiers=[
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ]
)


if __name__ == '__main__':
    setup(**settings)



from typing import Generic
from queue import Empty
from queue import Full

from .common import (
    T,
    OptInt,
    check_closing
)
from .queue import AbstractQueue


class SyncQueueProxy(Generic[T]):
    """Create a queue object with a given maximum size.

    If maxsize is <= 0, the queue size is infinite.
    """

    def __init__(
        self,
        parent: AbstractQueue[T]
    ) -> None:
        self._parent = parent

    @property
    def maxsize(self) -> int:
        return self._parent._maxsize

    @property
    def closed(self) -> bool:
        return self._parent.closed

    @check_closing
    def task_done(self) -> None:
        """Indicate that a formerly enqueued task is complete.

        Used by Queue consumer threads.  For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items
        have been processed (meaning that a task_done() call was received
        for every item that had been put() into the queue).

        Raises a ValueError if called more times than there were items
        placed in the queue.
        """

        with self._parent._all_tasks_done:
            unfinished = self._parent._unfinished_tasks - 1
            if unfinished <= 0:
                if unfinished < 0:
                    raise ValueError('task_done() called too many times')
                self._parent._all_tasks_done.notify_all()
                self._parent._loop.call_soon_threadsafe(
                    self._parent._finished.set)
            self._parent._unfinished_tasks = unfinished

    def join(self) -> None:
        """Blocks until all items in the Queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer thread calls task_done()
        to indicate the item was retrieved and all work on it is complete.

        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        with self._parent._all_tasks_done:
            while self._parent._unfinished_tasks:
                self._parent._all_tasks_done.wait()

    def qsize(self) -> int:
        """Return the approximate size of the queue (not reliable!)."""
        return self._parent._qsize()

    @property
    def unfinished_tasks(self) -> int:
        """Return the number of unfinished tasks."""
        return self._parent._unfinished_tasks

    @check_closing
    def put(
        self,
        item: T,
        block: bool = True,
        timeout: OptInt = None
    ) -> None:
        """Put an item into the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).
        """

        with self._parent._sync_not_full:
            if self._parent._maxsize > 0:
                if not block:
                    if self._parent._qsize() >= self._parent._maxsize:
                        raise Full
                elif timeout is None:
                    while self._parent._qsize() >= self._parent._maxsize:
                        self._parent._sync_not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    time = self._parent._loop.time
                    endtime = time() + timeout
                    while self._parent._qsize() >= self._parent._maxsize:
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise Full
                        self._parent._sync_not_full.wait(remaining)
            self._parent._put_internal(item)
            self._parent._sync_not_empty.notify()
            self._parent._notify_async_not_empty(threadsafe=True)

    @check_closing
    def get(
        self,
        block: bool = True,
        timeout: OptInt = None
    ) -> T:
        """Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        """

        with self._parent._sync_not_empty:
            if not block:
                if not self._parent._qsize():
                    raise Empty
            elif timeout is None:
                while not self._parent._qsize():
                    self._parent._sync_not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a non-negative number")
            else:
                time = self._parent._loop.time
                endtime = time() + timeout
                while not self._parent._qsize():
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise Empty
                    self._parent._sync_not_empty.wait(remaining)
            item = self._parent._get()
            self._parent._sync_not_full.notify()
            self._parent._notify_async_not_full(threadsafe=True)
            return item
